c$$$ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c$$$    Integrator for CMB anisotropy, CMB polarization and transfer functions
c$$$    Developed and maintained by Uros Seljak (useljak@princeton.edu) and 
c$$$    Matias Zaldarriaga (matiasz@sns.ias.edu)
c$$$
c$$$    Permission to use, copy, distribute and modify without fee or royalty at
c$$$    any tier, this software and its documentation, for any purpose 
c$$$    and without fee or royalty is hereby granted, provided that 
c$$$
c$$$    1)  you agree to comply with the following copyright 
c$$$    notice and statements, 
c$$$    including the disclaimer, and that the same appear on ALL copies of the 
c$$$    software and documentation, including modifications that you make for 
c$$$    internal use or for distribution:
c$$$     
c$$$    Copyright 1996, 1997, 1998 (versions up to 3.0) by 
c$$$    Harvard-Smithsonian Center for 
c$$$    Astrophysics and the Massachusetts Institute of Technology.  
c$$$    All rights reserved.
c$$$ 
c$$$    Copyright 1999, 2000 (version 3.0-4.0) by Princeton University
c$$$    and Insititute for Advanced Studies, Princeton. All rights
c$$$    reserved.
c$$$     
c$$$    THIS SOFTWARE IS PROVIDED "AS IS", AND INSTITUTIONS ABOVE MAKE NO 
c$$$    REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.  
c$$$    By way of example, but not limitation,
c$$$    THEY MAKE NO REPRESENTATIONS OR WARRANTIES OF 
c$$$    MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT 
c$$$    THE USE OF THE LICENSED SOFTWARE OR DOCUMENTATION WILL NOT INFRINGE 
c$$$    ANY THIRD PARTY PATENTS, COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS.   
c$$$     
c$$$    2)   Applications are limited to non-commercial and not-for-profit
c$$$    purposes.
c$$$
c$$$    3)   An appropriate acknowledgement is included in any publications 
c$$$    based on work conducted with usage of the CMBFAST package or any 
c$$$    of its derivatives.
c$$$
c$$$    Boltzmann integrator, initial conditions and recombination routines are 
c$$$    based on the COSMICS package of E. Bertschinger (written by Edmund 
c$$$    Bertschinger, Chung-Pei Ma and Paul Bode at MIT). See the LICENSE 
c$$$    file of the COSMICS distribution for restrictions on the modification 
c$$$    and distribution of this software.
c$$$
c$$$    Recombination routine RECFAST written by D. Scott (UBC), based on 
c$$$    calculations in the paper by Seager, Sasselov & Scott 
c$$$    (ApJ, 523, L1, 1999).
c$$$    See the license in recfast.f for further details regarding restrictions 
c$$$    on the modification and distribution of this software.
c$$$   
c$$$    Configure script, Makefile, fitsout.f, params.f, FITS and IARGC options 
c$$$    written by Matthias Bartelmann.  
c$$$cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                VERSION 4.0                           C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

	program driver
	implicit double precision(a-h,o-z)
c This is a driving routine that illustrates the use of the program.

	include 'cmbfast.inc'
        include 'cmbpar.inc'
	
c        parameter (nnmax=10)
c       const=7*pi**4/120.
	parameter (const=5.68219698d0,zeta3=1.20205690d0)
c     
	real*8 ztf(nnmax)
        character*80 filename
	character*80 filejlens
        character*80 parafile
cDE	writeway: common for output writing style: iwriteway=0, normal writing, 
cDE	iwriteway=1, cl_TT, cl_EE, cl_BB, cl_TE columns
	common /writeway/ iwriteway				!DE10
cDE     filenames: common for files containing  background and perturbation dynamics
        character*80 filebkg,fileprt                            !DE20
        common /filenames/ filebkg,fileprt                      !DE
	common /filej/ filejlens
	common /lens/ lensflag
	common /trlens/ntlens

c Output arrays temperature: clts, cltt ; e spectra:
c cles, clet ; b perturbation(only tensor contrubute):
c clbt ; cross correlation: clcs, clct.

        real*8 clts(l0max,nnmax),cltt(l0max,nnmax)
        real*8 cles(l0max,nnmax),clet(l0max,nnmax)
        real*8 clbs(l0max,nnmax),clbt(l0max,nnmax)
        real*8 clcs(l0max,nnmax),clct(l0max,nnmax)
        real*8 clkk(l0max,nnmax),cltk(l0max,nnmax)

	integer l(lmax),l0

	common /lvalues1/ l,l0,lmo
	save /lvalues1/

cDE     common lingerinc now contains: 
cDE     effective dark energy present and asimptotic equation of state wv, wvbeg, 
cDE     scalar field dark energy density parameter and equation of state today, omegaphif, wphif
	common /lingerinc/ omegab,omegac,omegav,wv,wvbeg,omegan
     &  ,omegak,h0,tcmb,yhe,annur,annunr,omegaphif,wphif          !DE30
cDE     common potpar and ggrpar with scalar field dark energy potential parameters
        common /potpar/ alphapot,amplipot,phif0,rkinpot,ipot      !DE40
        common /ggrpar/ betanmc,cnmc,omgjbd,sgnjbd,inmc           !DE50

	common /genparm/ grhom,grhog,grhor,adotrad,taurst,dtaurec
	common /initcase/ initfl
        common /reionization/zri,taurist,zristp,tauristp,rif,optdlss
        integer rcflag
        common /recfl/rcflag

        common /initialps/ an(nnmax),nn
        common /tensor/ ant(nnmax),rat(nnmax),itflag,lmaxt
        common /transfer/akmaxt,ztf,nlnkt,ict,ntf
	common /transfer2/ictpres

c       K split
	common /cutk/ aksplit, kcutflag

c	TIMING
c       Timing variables for testing purposes
c       Should be commented in released versions
c	real*4 etime,actual,timeprev,tarray(2)
c	external etime


c       TIMING
c       Enterning CMBFLAT
c	actual=etime(tarray)

	pi=dacos(-1.d0)						!DE60

#ifdef IARGC
        if (iargc() .eq. 0) then
           intflg = .true.
        else
           intflg = .false.
           if (iargc() .ne. 1) then
              print *, 'Usage: cmb [name of parameter file]'
              stop 1
           end if
           call getarg(1,parafile)
           call params(parafile)
        endif
#else 
        intflg = .true.
#endif

        if (intflg) then
           write(*,*)'CMB (0), transfer functions (1) or both (2):'
           write(*,*)'If you want the lensed Cls you will need (2)'
           read(*,*)ict
        else
           ict=p_ict
        endif

	if (ict.ne.0) then
          if (intflg) then
	   write(*,*)'If you want a high precision (1%)in' 
           write(*,*)'the transfer function at small scales'
	   write(*,*)'enter 1, otherwise 0 (The precision in the Cls' 
           write(*,*)'is not affected by this choice but the code'
	   write(*,*)'will become slower).'
	   read(*,*)ictpres
           else
              ictpres=p_prc
           endif
	end if
	if (ict.ne.1) then
          if (intflg) then
	   write(*,*)'Value of lmax, ketamax (e.g. 1500 3000)'
	   write(*,*)'Remember to be consistent with the file'
	   write(*,*)'in the flat case.'
	   read(*,*)lmoin,akmax0
          else
           lmoin=p_lmx
           akmax0=p_emx
          endif

	   call initlval(lmoin,akmax0)
	endif

	
       if (ict.ne.0) then
c output at z=10 saves 50% of time; for CDM models tf remains 
c unchanged between z=0 and z=10; for MDM, lambda, dark energy, 
c non-flat models one should integrate until z=0
      if (intflg) then
        write(*,*)'Enter tf kmax (h/Mpc), # of k per log. int. (5,5)' 
        read(*,*)akmaxt,nlnkt
        write(*,*)'Enter number and redshifts of the output: (1,0)'
        write(*,*)'If more than one tf is requested the redshifts'
        write(*,*)'have to be specified in decreasing order.'
        read(*,*)ntf,(ztf(i),i=1,ntf)
      else
        akmaxt=p_kmx
        nlnkt=p_nkl
        ntf=p_ntf
        do itf=1,ntf
           ztf(itf)=p_ztf(itf)
        enddo
      endif

       do itf=1,ntf
        if (intflg) then
         write(*,*)'Enter',itf,'. output transfer function file'
         read(*,'(a80)') filename
        else
            filename=p_ftf(itf)
        endif
        open(unit=12+itf,file=filename,form='formatted'
     <,status='unknown')
        enddo
        if (intflg) then
cDE     text below changed accordingly to dark energy cosmologies
        write(*,*)'output file consists of k(h/Mpc) and tf for '
        write(*,*)'CDM, baryons, photons, massless neutrinos,'
        write(*,*)'massive neutrinos when applicable,'
        write(*,*)'scalar field dark energy when applicable'     !DE70
        endif
       endif
c  Read initial parameters.

CC1 If you want to make loops over models remove the
CC1 following questions ad replace them by the loop giving
CC1 values to all the variables,
CC1 omegab,omegac,omegav,wv,wvbeg,omegan,omegaphif,h0,tcmb,yhe,annunr, !DE75
CC1 riflag,optdlss or zri,rif,itflag,nn,an,ant (if tensors
CC1 are wanted),initfl.
CC1 Remember that for nonflat models massive neutrinos, gravity 
CC1 waves or initial conditions other than isentropic are not 
CC1 available. 

1       if (intflg) then                                                                              !DE80
	   write(*,*)                                                                                 !DE
cDE     read scalar field density parameter 
     &  'Enter Omega_b, Omega_c, Omega_v, Omega_nu, Omega_phi', 
     &  '[.044 .226 0.730 0. 0.]:'                                                                     !DE
	read(*,*) omegab,omegac,omegav,omegan,omegaphif                                               !DE
cDE	read dark energy equation of state
		if(omegav.ne.0.0d0)then                                                               !DE
        write(*,*)'Enter dark energy equation of state modeled as' 
	write(*,*)'w=w_0+(w_infty-w_0)(a-1):'					                      !DE
        write(*,*)'Enter effective dark energy'
        write(*,*)'present equation of state [-1.]:'                                                  !DE
        read(*,*) wv							                              !DE
        write(*,*)'Enter effective dark energy'
        write(*,*)'asimptotic equation of state [-1.]:'	                                              !DE
        read(*,*) wvbeg							                              !DE
 		        if((wv.ne.-1.d0).or.(wvbeg.ne.-1.d0)) 
     &  print*,'Effective dark energy model 
     &  (dynamical cosmological constant) selected.'                                                  !DE
		        else                                                                          !DE
			wv=-1.d0                                                                      !DE
			wvbeg=-1.d0                                                                   !DE
		end if                                                                                !DE
cDE     Scalar field dark energy parameters
cDE     the generalised Lagrangian is (1/2)fnmc(phif)R-(1/2)onmc(phif)*phif_;mu*phif^;mu-V(phif)+L_matter ;
cDE     ipot=0 nullifies the potential, V=0 ;
cDE     inmc=0 nullifies non-minimal coupling, fnmc=0, onmc=1 ;
cDE     ipot=1 makes inverse power law potential V(phif)=M^4/(phif/phif0)^alphapot and field in tracking regime ;
cDE     ipot=2 makes SUGRA potential V(phif)=(M^4/(phif/phif0)^alphapot)exp(4pi(phif^2-phif0^2)) and field in tracking regime ;
cDE     inmc=1 makes fnmc=(1+8pi*cnmc*(phif^betanmc-phif0^betanmc)*R/(8piG), today f(phif0)=1/(8piG), onmc=1 ;
cDE     now initializes variables: 
        alphapot=0.                                                                                   !DE
        amplipot=0.                                                                                   !DE
        ipot=0                                                                                        !DE
        phif0=1.                                                                                      !DE
        betanmc=0.                                                                                    !DE
        cnmc=0.                                                                                       !DE
        inmc=0                                                                                        !DE
        omgjbd=1.d15                                                                                  !DE
        sgnjbd=1.                                                                                     !DE
		if(omegaphif.gt.0.)then                                                               !DE
	write(*,*)'Scalar field dark energy models selected.'                                         !DE
        write(*,*)'Enter scalar field dark energy 
     &  equation of state [-0.9]:'                                                                    !DE
        read(*,*) wphif							                              !DE
1967    print*,'Enter scalar field dark energy potential index 
     &  [1, tracking inverse power law; 2, tracking SUGRA]:'                                          !DE
        read (*,*) ipot                                                                               !DE
			if((ipot.eq.1).or.(ipot.eq.2))then                                            !DE
cDE     Setting initial potential exponent -2, to be adjusted numerically to match 
cDE     the equation of state today
        alphapot=-2.d0 								                      !DE
cDE     Setting initial potential amplitude at the Planck mass, to be fixed numerically later to match 
cDE     the required Omega_phi today
        amplipot=1.d0   						                              !DE
cDE     Setting initial scalar field value at one tenth of the Planck mass, to be fixed numerically later 
cDE     to match the required Omega_phi today 
        phif0=1.d-1		                                                                      !DE
cDE	Setting initial ratio between kinetic and potential energy densities equal to 0
cDE	Tracking trajectories are greatly stable with respect to variation of this parameter
	rkinpot=0.                                                                                    !DE
cDE        write (*,*)'Enter initial kinetic versus potential energy ratio [0.]:'                     !DE
cDE        read (*,*) rkinpot                                                                         !DE
cDE     reading file names for background and perturbation dynamics
        print*,'Enter file names to output background dynamics'                                       !DE
	print*,'and pertutbation amplitude [bkg.dat,prt.dat]:'                                        !DE
cDE	filebkg='bkg.dat'                                                                             !DE
cDE     fileprt='prt.dat'                                                                             !DE
        read(*,*) filebkg                                                                             !DE
        read(*,*) fileprt                                                                             !DE
	open(1,file=fileprt,status='unknown')                                                         !DE
	close(1)                                                                                      !DE
			else if(ipot.ge.3)then                                                        !DE
        print*,'sorry, only tracking inverse power and'                                               !DE
        print*,'SUGRA potentials supported.'                                                          !DE
        go to 1967                                                                                    !DE
			end if                                                                        !DE
1968    write (*,*) 'Enter scalar field dark energy non-minimal coupling index 
     &  [0 for minimal coupling, 1/8piG, 1 for 1/8piG+xi*phi^2]:'                                     !DE
        read (*,*) inmc                                                                               !DE
		        if(inmc.eq.1)then                                                             !DE
cDE     Setting coupling exponent 2
        betanmc=2. 								                      !DE
        print*,'Enter omegaJBD 
     &  [4.e4, present lower limit from Cassini measurements]:'	                                      !DE
        read (*,*) omgjbd                                                                             !DE
        print*,'Enter non-minimal coupling constant sign 
     &  [1. for xi>0. -1. for xi<0]:'                                                                 !DE
        read (*,*) sgnjbd                                                                             !DE
cDE     Setting initial value for the non-minimal coupling constant, to be adjusted numerically later to
cDE     match the required OmegaJBD today
        cnmc=sgnjbd/dsqrt(32.d0*pi*omgjbd*phif0**2)                                                   !DE
			else if(inmc.ge.2)then                                                        !DE
        print*,'sorry, only non-minimal coupling supported'                                           !DE
        go to 1968                                                                                    !DE
			end if                                                                        !DE
	        else if(omegaphif.eq.0.0)then                                                         !DE
        wphif=-1.0                                                                                    !DE
        print*,'Enter file name to output pertutbation 
     &  amplitude [prt.dat]:'                                                                         !DE
	filebkg='bkg.dat'                                                                             !DE
cDE     fileprt='prt.dat'                                                                             !DE
cDE     read(*,*) filebkg                                                                             !DE
        read(*,*) fileprt                                                                             !DE
	open(1,file=fileprt,status='unknown')                                                         !DE
	close(1)                                                                                      !DE
		else if(omegaphif.lt.0.)then                                                          !DE
	write(*,*) 'Sorry, only positive densities                                                    !DE
     &  for scalar field are allowed.'                                                                !DE
	stop                                                                                          !DE
		end if                                                                                !DE
        else                                                                                          !DE
           omegab=p_par(1)                                                                            !DE
           omegac=p_par(2)                                                                            !DE
           omegav=p_par(3)                                                                            !DE
           omegan=p_par(4)                                                                            !DE
           omegaphif=p_par(5)                                                                         !DE
           wv=p_par(6)                                                                                !DE
           wvbeg=p_par(7)                                                                             !DE
           wphif=p_par(8)                                                                             !DE
           alphapot=0.                                                                                !DE
           amplipot=0.                                                                                !DE
           phif0=1.                                                                                   !DE
           rkinpot=0.                                                                                 !DE
           ipot=int(pot_par(1))                                                                       !DE
	           if((ipot.eq.1).or.(ipot.eq.2))then                                                 !DE
cDE        Setting initial potential exponent -2, to be adjusted numerically to match 
cDE        the equation of state today
           alphapot=-2. 							                      !DE
cDE        Setting initial potential amplitude at the Planck mass, to be fixed numerically later to match 
cDE        the required Omega_phi today
           amplipot=1.							                              !DE
cDE        Setting initial scalar field value at one tenth of the Planck mass, to be fixed numerically later 
cDE        to match the required Omega_phi today 
           phif0=0.1		                                                                      !DE
cDE	   Setting initial ratio between kinetic and potential energy densities equal to 0
cDE	   Tracking trajectories are greatly stable with respect to variation of this parameter
	   rkinpot=0.                                                                                 !DE
cDE           write (*,*)'Enter initial kinetic versus potential energy ratio:'                       !DE
cDE            read (*,*) rkinpot                                                                     !DE
cDE     reading file names for background and perturbation dynamics
        print*,'Enter file names to output background dynamics'                                       !DE
	print*,'and pertutbation amplitude [bkg.dat,prt.dat]:'                                        !DE
cDE	filebkg='bkg.dat'                                                                             !DE
cDE     fileprt='prt.dat'                                                                             !DE
        read(*,*) filebkg                                                                             !DE
        read(*,*) fileprt                                                                             !DE
	open(1,file=fileprt,status='unknown')                                                         !DE
	close(1)                                                                                      !DE
	           else if(ipot.ge.3) then                                                            !DE
           print*,'sorry, only tracking inverse power and'                                            !DE
           print*,'SUGRA potentials supported.'                                                       !DE
           stop                                                                                       !DE
		   end if                                                                             !DE
           betanmc=0.                                                                                 !DE
           cnmc=0.                                                                                    !DE
           omgjbd=1.d15                                                                               !DE
           sgnjbd=1.                                                                                  !DE
	   inmc=int(nmc_par(1))                                                                       !DE
           omgjbd=nmc_par(2)                                                                          !DE
           sgnjbd=nmc_par(3)                                                                          !DE
        inmc       = inmc_p                                                                           !DE
		if(inmc.eq.1) then                                                                    !DE
cDE     Setting coupling exponent 2
        betanmc=2. 								                      !DE
cDE     Setting initial value for the non-minimal coupling constant, to be adjusted numerically later to
cDE     match the required OmegaJBD today
        cnmc=sgnjbd/dsqrt(32.d0*pi*omgjbd*phif0**2)                                                   !DE
	        else if(inmc.ge.2) then                                                               !DE
        print*,'sorry, only non-minimal coupling supported.'                                          !DE
        stop                                                                                          !DE
		end if                                                                                !DE
        endif                                                                                         !DE
	omega=omegab+omegac+omegav+omegan+omegaphif                                                   !DE

	omegak=1.0d0-omega

cDE     avoid omegak and omegaphi models 
        if (abs(omegak).ge.1.e-3.and.omegaphif.ne.0.0) then                                           !DE90
        print*,'scalar field dark energy is allowed only 
     &  in flat universe.'                                                                            !DE
        stop                                                                                          !DE
        end if                                                                                        !DE

2       if (intflg) then
	   write(*,*)
     2    'Enter H0, Tcmb, Y_He, N_nu(massless), N_nu(massive)',
     3    '(e.g. 72. 2.726 0.24 3. 0.)'                                                               !DE95
	read(*,*) h0,tcmb,yhe,annur,annunr
        else
cDE     parameter indexes changed to account for wv, wvbeg and omegaphif
           h0=p_par(9)                                                                                !DE100
           tcmb=p_par(10)                                                                             !DE
           yhe=p_par(11)                                                                              !DE
           annur=p_par(12)                                                                            !DE
           annunr=p_par(13)                                                                           !DE
        endif
        akmaxt=akmaxt*(h0/100.0d0)
	if (h0.lt.25.d0.or.h0.gt.100.d0) then
	  write(*,*)
     2      '  Warning: H0 has units of km/s/Mpc.  Your value is weird.'
	end if
	if (tcmb.lt.2.7d0.or.tcmb.gt.2.8d0) then
	  write(*,*)
     2      '  Warning: Tcmb has units of K.  Your value is weird.'
	end if

	if (yhe.lt.0.2d0.or.yhe.gt.0.3d0) then
	  write(*,*)
     2      '  Warning: Y_He is the Helium fraction of baryons.',
     3      '  Your value is weird.'
	end if
        if (annunr.lt.0.or.annunr.gt.3.1) then
          write(*,*)
     2      'Warning: N_nu(massive) is strange'
          write(*,*) '  Illegal choice.  Try again:'
          if (intflg) go to 2
          stop 1
        end if
        if (annur.lt.0.or.annur.gt.3.1) then
          write(*,*)
     2      'Warning: N_nu(massless) is strange'
          write(*,*) '  Illegal choice.  Try again:'
          if (intflg) go to 2
          stop 1
        end if
        if (annunr.lt.1.and.omegan.gt.0.0) then
          write(*,*)
     2      'Warning: N_nu(massive) should be 1, 2, or 3',
     3      'For non zero omegan'
	   write(*,*) '  Illegal choice.  Try again:'
          if (intflg) go to 2
          stop 1
        end if

        if (intflg) then
        write(*,*)  'Enter 0 for Peebles recombination 
     &  (required for dark energy models), or 1 for recfast'                                         !DE105
        read(*,*) rcflag
        else
           rcflag=p_rcf
        endif

cDE	recfast uses now dtauda; it needs the constants below, which have been slightly anticipated; 
cDE	moreover, its call in scalar field cosmologies has been moved in cmbflat, after 
cDE	the background table has been calculated

c  Evaluate general cosmology constants.

c  grho gives the contribution to the expansion rate from: (g) photons,
c  (r) one flavor of relativistic neutrino (2 degrees of freedom),
c  (m) nonrelativistic matter (for Omega=1).  grho is actually
c  8*pi*G*rho/c^2 at a=1, with units of Mpc**(-2).
c  a=tau(Mpc)*adotrad, with a=1 today, assuming 3 neutrinos.
c  (Used only to set the initial conformal time.)
	grhom=3.3379d-11*h0*h0
	grhog=1.4952d-13*tcmb**4
	grhor=3.3957d-14*tcmb**4

c  adotrad gives the relation a(tau) in the radiation era:
	adotrad=2.8948d-7*tcmb*tcmb

        if ((ipot.eq.0).and.(inmc.eq.0))then                    	                             !DE
           if (rcflag.eq.1) then
           write(*,*) 'Calling recfast'
           call recfast(omegab,omegac,omegav,h0,tcmb,yhe)
           write(*,*)'recfast succesfully completed'
           endif
        endif                                                   	                             !DE105.5

	if (intflg) then
	   write (*,*) 'Enter 0 for no reionization,'
	   write (*,*) '1 for specified'
	   write (*,*) 'optical depth to lss (xe=1),'
	   write (*,*) '2 for specified redshift and xe [1]'                                         !DE106
	   read (*,*) riflag
	else
	   riflag=p_rif
	endif

	zri=0.0d0
	rif=0.0d0
	optdlss=0.0d0
	if (riflag.eq.1) then 
           if (intflg) then
	   write (*,*) 'Enter optical depth to lss [0.11]'
	   read (*,*) optdlss
           else
              optdlss=p_opd
           endif
	   rif=1.0d0
	end if

	if (riflag.eq.2) then
           if (intflg) then
	   write (*,*)'Enter redshift, ionization fraction(e.g. 50 0.2)'
	   read (*,*) zri,rif
           else
              zri=p_red
              rif=p_frc
           endif
	   zristp=0.07d0*zri-1.0d0
	   if (zristp.lt.0.0) zristp=0.0d0
	end if

	if (ict.ne.1) then
           if (intflg) then
	      write (*,*) 'Enter 0 for scalar alone,' 
	      write(*,*) '1 for tensor+scalar 2 for tensors alone'
	      write(*,*)'3 for scalar (k<k*) and 4 for scalar (k>k*) [0]'                             !DE107
	      read(*,*)itflag
           else
              itflag=p_tfl
           endif

c       K splitting
	   aksplit=1.0d0
	   kcutflag=0
	   if (itflag.eq.3) then
	      itflag=0
	      kcutflag=1
	   end if
	   if (itflag.eq.4) then
	      itflag=0
	      kcutflag=-1
	   end if
	   
	   if (itflag.ne.2) then	 
              if (intflg) then
	      write(*,*)'number and values of scal. spec. ind. n [1,0.96]'                            !DE108
	      read(*,*)nn,(an(i),i=1,nn)
              else
                 nn=p_nsm
                 do i=1,nn
                    an(i)=p_scl(i)
                 enddo
              endif
	   end if
	   if (itflag.eq.1) then 
              if (intflg) then
	      write(*,*)
     2 'Tensor spectral index given by nt=ns-1 (0) or different (1)'
	      read(*,*)itn
              else
                 itn=p_itn
              endif

	      if (itn.eq.0) then
		 do 123 in=1,nn
		    ant(in)=an(in)-1
 123		 continue
                 if (.not.intflg) then
                    do in=1,nn
                       p_ten(in)=ant(in)
                    enddo
                 endif

	      else
                 if (intflg) then
		 write(*,*)'values of tensor spectral indexes:'
		 read(*,*)(ant(in),in=1,nn)
                 else
                    do in=1,nn
                       ant(in)=p_ten(in)
                    enddo
                 endif
	      endif
	   endif
	   if (itflag.eq.2) then
              if (intflg) then
	      write(*,*)'number and values of tens. spec. ind. (1,0)'
	      read(*,*)nn,(ant(in),in=1,nn)
           else
              nn=p_nsm
              do in=1,nn
                 ant(in)=p_ten(in)
              enddo
           endif
	      do in=1,nn
		 rat(in)=1.0d0
	      enddo
	   end if
           if (itflag.eq.1) then
              if (intflg) then
	      write(*,*)'ratio of tensor to scalar quadrupole given by' 
	      write(*,*)'7(1-n_S) (0) or different (1):'
	      read(*,*)irt
              else
                 irt=p_irt
              endif

	      if (irt.eq.0) then
		 do in=1,nn
		    rat(in)=7.0d0*(1.0d0-an(in))
		 enddo
                 if (.not.intflg) then
                    do in=1,nn
                       p_rat(in)=rat(in)
                    enddo
                 endif
	      else
                 if (intflg) then
		 write(*,*)'values of T/S:'
		 read(*,*)(rat(in),in=1,nn)
              else
                 do in=1,nn
                    rat(in)=p_rat(in)
                 enddo
	      endif
	      endif
	   end if
	   if (itflag.eq.0) then
	      do in=1,nn
		 rat(in)=1.0d0
	      enddo
           endif 
	   
	   lensflag=0
	   if ((itflag.ne.2).and.(ict.eq.2)) then
              if (intflg) then
	      write(*,*)'Enter 0 unlensed Cls only,'
	      write(*,*)'1 lensed Cls, linear evolution'
	      write(*,*)'2 lensed Cls, non-linear evolution [1]'                                      !DE109
	      if (itflag.eq.1) then
		 write(*,*)'ONLY SCALAR Cls ARE LENSED'
	      end if
	      read(*,*)lensflag
              else
                 lensflag=p_glf
              endif
	   end if
	   if ((lensflag.ne.0).and.(ict.eq.0)) then
	      write(*,*)'You did not request the transfer function'
	      write(*,*)'calculation needed to do the lensing'
	      write(*,*)'you will have to start again'
	      stop
	   else
c The last requested transfer function will be used for the lensing	      
	      ntlens=ntf
	   end if

cDE        check way of writing outputs 
	   write(*,*) 'Please, enter 1 for sharp writing, 0 otherwise: ' !DE110
           read(*,*) iwriteway                                           !DE
	   if (iwriteway.eq.0) then 	                                 !DE
	   if (itflag.ne.2) then
              if (intflg) then
	      write(*,*)'Enter output filename for SCALAR cl'
	      if (lensflag.ne.0) then
		 write(*,*)'If lensing was requested this will'
		 write(*,*)'be the unlensed power spectrum'
	      end if
	      read(*,'(a80)')filename
              else
                 filename=p_fcl
              endif

	      open(unit=8,file=filename,
     &	      	   status='unknown',form='formatted')
	      rewind 8

	      if (lensflag.ne.0) then
                 if (intflg) then
		 write(*,*)'Enter output filename for LENSED SCALAR cl'
		 read(*,'(a80)')filename
                 else
                    filename=p_fgl
                 endif
		 open(unit=10,file=filename,
     &            status='unknown',form='formatted')
		 rewind 10

                 if (intflg) then
		 write(*,*)'Enter file with bessel functions' 
                 write(*,*)' for lensing calculation (jlens.dat)'
		 read(*,'(a80)')filejlens
                 else
                    filejlens=p_jgl
                 endif
	      end if
	   end if

	   if (itflag.ne.0) then
              if (intflg) then
	      write(*,*)'Enter output filename for TENSOR cl'
	      read(*,'(a80)')filename
              else
                 filename=p_tcl
              endif
	      open(unit=9,file=filename,
     &      	   status='unknown',form='formatted')
	      rewind 9
	   end if
	  else if (iwriteway.ne.0) then               	     !DE120
	      write(*,*)'Sharp writing mode selected.'       !DE
	      write(*,*)'Enter output filename for cl'       !DE 
	      read(*,'(a80)')filename                        !DE 
	      open(unit=8,file=filename,                     !DE 
     &	      	   status='unknown',form='formatted')        !DE
	      rewind 8                                       !DE
          end if                                             !DE  

	else
	   itflag=0
	   nn=0
	   lensflag=0
	end if

#ifdef FITS
        if (intflg) then
           write (*,*) 'Enter FITS file name'
           read (*,'(a80)') p_fts
        end if
#endif

	if (itflag.ne.2) then
	   if (abs(omegak).gt.1.0d-3) then
              if (intflg) then
	      write (*,*) 'Enter initial conditions'
	      write (*,*) '1= Isentropic (adiabatic)'
	      write (*,*) '2= Isocurvature CDM'
	      write (*,*) '3= Isocurvature baryon'
	      write (*,*) '4= Isocurvature seed conditions'
	      read (*,*) initfl
              else
                 initfl=p_inf
              endif
	   else
              if (intflg) then
	      write (*,*) 'Enter initial conditions'
	      write (*,*) '1= Isentropic (adiabatic)'
	      write (*,*) '2= Isocurvature CDM'
	      write (*,*) '3= Isocurvature baryon'
	      write (*,*) '4= Isocurvature seed conditions'
cDE   isocurvature in scalar field cosmology if requested 
	      if((ipot.ne.0).or.(inmc.ne.0))then                          !DE130
	      write(*,*)'5=Isocurvature CDM in scalar field cosmology'    !DE
	      write(*,*)'6=Isocurvature baryon in scalar field cosmology' !DE
	      write(*,*)'7=Isocurvature scalar'                           !DE
	      end if                                                      !DE
	      read (*,*) initfl
              else
                 initfl=p_inf
              endif
	   end if
	else
	   initfl=1
	end if

        if (intflg) then
           p_frc=rif
           p_glf=lensflag
           p_ict=ict
           p_inf=initfl
           p_irt=irt
           p_itn=itn
           p_kmx=akmaxt
           p_nkl=nlnkt
           p_nsm=nn
           p_ntf=ntf
           p_opd=optdlss
           p_par(1)=omegab
           p_par(2)=omegac
           p_par(3)=omegav
           p_par(4)=omegan
cDE        account for scalar field density parameter and effective dark energy 
cDE        equation of state in parameters
           p_par(5)=omegaphif                  !DE140
           p_par(6)=wv                         !DE
           p_par(7)=wvbeg                      !DE
           p_par(8)=wphif                      !DE
           p_par(9)=h0
           p_par(10)=tcmb
           p_par(11)=yhe
           p_par(12)=annur
           p_par(13)=annunr
           p_prc=ictpres
           p_rat(1)=rat(1)
           p_red=zri
	   p_rcf=rcflag
           p_rif=riflag
           p_scl(1)=an(1)
           p_ten(1)=ant(1)
           p_ten(1)=ant(1)
           p_tfl=itflag
           p_ztf(1)=ztf(1)
        end if
        if (.not.intflg) call getpar(.false.)

c     TIMING
c	timeprev=actual
c	actual=etime(tarray)
c	write(50,*)actual-timeprev,' END INPUT, NOW MAIN'
c	write(50,*)'PARAMETERS'
c	write(50,*)omegab,omegac,omegav,wv,wvbeg,omegan,omegaphif,wphif !DE145 
c	write(50,*)h0,tcmb,yhe,annur,annunr
c	write(50,*)riflag,zri,rif,optdlss

c call main subroutine

CC2 If you are interested in constructing a driver for
CC2 only the flat or nonflat models code, leave the appropiate 
CC2 piece of following if statement. Then enter the appropiate
CC2 parameters for that case. 

	do in=1,nnmax
	   do il=1,l0max
	      clts(il,in)=0.0d0
	      cltt(il,in)=0.0d0
	      cles(il,in)=0.0d0
	      clet(il,in)=0.0d0
	      clbs(il,in)=0.0d0
	      clbt(il,in)=0.0d0
	      clcs(il,in)=0.0d0
	      clct(il,in)=0.0d0
	      clkk(il,in)=0.0d0
	      cltk(il,in)=0.0d0
	   end do
	end do


       if (abs(omegak).le.1.0d-3) then
CC1 initjl reads the tables of spherical Bessel functions
CC1 so if you are making a loop over models you should
CC1 move this call outside the loop.
	  if (ict.ne.1) then

	     if (intflg) then
		write(*,*)'Enter input filename for jl'
		read(*,'(a80)')filename
	     else
		filename=p_fjl
	     endif
	     open(unit=11,file=filename
     1         ,status='unknown',form='unformatted')
	     rewind 11
	     call initjl(filename)
	  end if
          call cmbflat(clts,cltt,cles,clet,
     1         clbt,clcs,clct,clkk,cltk)
	else
	   if (ict.ne.1) then
	      if (intflg) then
		 write(*,*)'Enter input filename for ujl'
		 read(*,'(a80)')filename
	      else
		 filename=p_ujl
	      endif
	      call initujl0(filename)
	   end if
	   call cmbopen(clts,cltt,cles,clet,
     1          clbt,clcs,clct,clkk,cltk)
	end if

c       TIMING
c	timeprev=actual
c	actual=etime(tarray)
c	write(50,*)actual-timeprev,' END MAIN, NOW COBE'



c For many purposes it is useful not to call 
c COBE normalization so that there is a direct relation 
c between the initial power spectrum and the CMB. 

#ifdef UNNORM
c       Not calling COBE normalization
#else
	call COBEnormalize(clts,cltt,cles,clet,clbt,
     &  clcs,clct,clkk,cltk)
#endif

c       TIMING
c	timeprev=actual
c	actual=etime(tarray)
c	write(50,*)actual-timeprev,' END COBE, NOW OUTPUT'


	call output(clts,cltt,cles,clet,clbt,
     &  clcs,clct,itflag,lmo)

c       TIMING
c	timeprev=actual
c	actual=etime(tarray)
c	write(50,*)actual-timeprev,' END OUTPUT, NOW LENSING'

	
	if (lensflag.ne.0) then
	   call lensing(clts,cles,clbs,clcs,clkk,cltk)
	end if

	if (itflag.ne.2) close(8)
	if (itflag.ne.0) close(9)

#ifdef FITS
        call fitsout(clts,cles,clbt,clcs,itflag,lmo)
#endif


c       TIMING
c	timeprev=actual
c	actual=etime(tarray)
c	write(50,*)actual-timeprev,' END LENSING, NOW END'

CC1 End your Loop over models here.
       stop
       end
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
