C Integrator for CMB anisotropy, CMB polarization and transfer functions
C Developed by Uros Seljak (useljak@princeton.edu)
C and Matias Zaldarriaga (matiasz@sns.ias.edu).
C See the LICENSE file for restrictions on the use, modification and
C distribution of this software.

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

cDE     includes dark energy density parameter and equation of state 
cDE     in the arguments
	subroutine run_cmbfast(
     &          omega_b,omega_c,omega_v,omega_n,omega_p,w_v,w_vbeg,
     &          w_p,ipot_p,inmc_p,omgjbd_p,sgnjbd_p,
     &		H_0,t_cmb,Y_he,a_nu_r,a_nu_nr,
     &		tau,num_slope,slope_n,ntype,l_max,aketamax,outfile)            !DE10

	implicit double precision(a-h,o-z)
c This is a driving routine that illustrates the use of the program.

cDE     includes omega scalar field and effective dark energy equation of state 
cDE     in the variables declared
	real omega_b,omega_c,omega_v,omega_n,omega_p,w_v,w_p                   !DE20
	real H_0,t_cmb,Y_he,a_nu_r,a_nu_nr
	real aketamax

	include 'cmbfast.inc'

c       parameter (nnmax=10)

	real tau, slope_n(nnmax)
	integer ntype, type ! 0 = scalar, 1 = tensor (does EITHER one OR the other)
			! slope_n refers to either scalar or tensor index, depending on type.
	integer l_max	! l-value up to which C_l should be computed
	integer num_slope ! number of ns to be calculated must be less than 20

	
c       const=7*pi**4/120.
	parameter (const=5.68219698d0,zeta3=1.20205690d0)
c     
	real*8 ztf(nnmax)
        character*80 filename, outfile
	character*80 filejlens
cDE	writeway: common for output writing style: iwriteway=0, normal writing, 
cDE	iwriteway=1, cl_TT, cl_EE, cl_BB, cl_TE columns
	common /writeway/ iwriteway				               !DE30
	common /filej/ filejlens
	common /lens/ lensflag
	common /trlens/ntlens

c Output arrays temperature: clts, cltt ; e spectra:
c cles, clet ; b perturbation(only tensor contrubute):
c clbt ; cross correlation: clcs, clct.

        real*8 clts(l0max,nnmax),cltt(l0max,nnmax)
        real*8 cles(l0max,nnmax),clet(l0max,nnmax)
        real*8 clbs(l0max,nnmax),clbt(l0max,nnmax)
        real*8 clcs(l0max,nnmax),clct(l0max,nnmax)
        real*8 clkk(l0max,nnmax),cltk(l0max,nnmax)

	integer l(lmax),l0

	common /lvalues1/ l,l0,lmo
	save /lvalues1/

cDE     common lingerinc now contains: 
cDE     effective dark energy present and asimptotic equation of state wv, wvbeg, 
cDE     scalar field dark energy density parameter and equation of state today, omegaphif,wphif
	common /lingerinc/ omegab,omegac,omegav,wv,wvbeg,omegan
     &  ,omegak,h0,tcmb,yhe,annur,annunr,omegaphif,wphif          !DE40
cDE     common potpar and ggrpar with scalar field dark energy potential parameters
        common /potpar/ alphapot,amplipot,phif0,rkinpot,ipot      !DE50
        common /ggrpar/ betanmc,cnmc,omgjbd,sgnjbd,inmc           !DE60

	common /genparm/ grhom,grhog,grhor,adotrad,taurst,dtaurec
	common /initcase/ initfl
        common /reionization/zri,taurist,zristp,tauristp,rif,optdlss
        integer rcflag
        common /recfl/rcflag


        common /initialps/ an(nnmax),nn
        common /tensor/ ant(nnmax),rat(nnmax),itflag,lmaxt
        common /transfer/akmaxt,ztf,nlnkt,ict,ntf
	common /transfer2/ictpres

c       K split
	common /cutk/ aksplit, kcutflag

c	TIMING
c       Timing variables for testing purposes
c       Should be commented in released versions
c	real*4 etime,actual,timeprev,tarray(2)
c	external etime

c       TIMING
c       Enterning CMBFLAT
c	actual=etime(tarray)
c	timeinit=actual

c       FOR PROJECT WITH MAX
        
	if ((ntype.ne.0).and.(ntype.ne.1).and.(ntype.ne.2)
     $  	.and.(ntype.ne.3)) then
	   write(*,*)'Invalid type'
	   stop
	end if

c       KSPLIT
	aksplit=1.0d0
	kcutflag=0
	if (ntype.eq.0) then
	   type = 0
	   kcutflag=0
	end if
	if (ntype.eq.1) then
	   type = 1
	   kcutflag=0
	end if
	if (ntype.eq.2) then
	   type = 0
	   kcutflag=1
	end if
	if (ntype.eq.3) then
	   type = 0
	   kcutflag=-1
	end if

c       TIMING
c       Enterning CMBFLAT
c	actual=etime(tarray)

c       THIS DRIVER IS JUST FOR COMPUTING CMB
c	write(*,*)'CMB (0), transfer functions (1) or both (2):'
c	write(*,*)'If you want the lensed Cls you will need (2)'
c	read(*,*)ict
	ict=0

	if (ict.ne.0) then
	   write(*,*)'If you want a high precision (1%)in' 
           write(*,*)'the transfer function at small scales'
	   write(*,*)'enter 1, otherwise 0 (The precision in the Cls' 
           write(*,*)'is not affected by this choice but the code'
	   write(*,*)'will become slower).'
	   read(*,*)ictpres
	end if
	if (ict.ne.1) then
	   if (aketamax.eq.0.0) then
	      akmax0=200.0d0+2.0d0*dble(l_max)
	      akmax0=min(akmax0,5000.d0)
	   else
	      akmax0=dble(aketamax)
	   end if
	   call initlval(l_max,akmax0)
	endif

	
       if (ict.ne.0) then
c output at z=10 saves 50% of time; for CDM models tf remains 
c unchanged between z=0 and z=10; for MDM, lambda, open models
c one should integrate until z=0
        write(*,*)'Enter tf kmax (h/Mpc), # of k per log. int. (5,5)' 
        read(*,*)akmaxt,nlnkt
        write(*,*)'Enter number and redshifts of the output: (1,0)'
        write(*,*)'If more than one tf is requested the redshifts'
        write(*,*)'have to be specified in decreasing order.'
        read(*,*)ntf,(ztf(i),i=1,ntf)
        do itf=1,ntf
         write(*,*)'Enter',itf,'. output transfer function file'
         read(*,'(a80)') filename
         open(unit=12+itf,file=filename,form='formatted'
     <,status='unknown')
        enddo
cDE     text below changed accordingly to dark energy cosmologies
        write(*,*)'output file consists of k(h/Mpc) and tf for '
        write(*,*)'CDM, baryons, photons, massless neutrinos'
        write(*,*)'and massive neutrinos when applicable,'
        write(*,*)'scalar field dark energy when applicable'     !DE70
       endif
 
c  Read initial parameters.

CC1 If you want to make loops over models remove the
CC1 following questions ad replace them by the loop giving
CC1 values to all the variables,
CC1 omegab,omegac,omegav,wv,wvbeg,omegan,omegaphif,h0,tcmb,yhe,annunr, !DE75
CC1 riflag,optdlss or zri,rif,itflag,nn,an,ant (if tensors
CC1 are wanted),initfl.
CC1 Remember that for open models massive neutrinos, gravity 
CC1 waves or initial conditions other than isentropic are not 
CC1 available. 


c1	write(*,*)
cDE     modified text and inputs to account for the effective dark energy equation of state as well as 
cDE     the amount of scalar field dark energy density and equation of state
cDE     2  'Enter Omega_b, Omega_c, Omega_v, Omega_nu, Omega_phi (e.g. .044 .226 0.730 0. 0.)'
cDE	read(*,*) omegab,omegac,omegav,omegan,omegaphif
	omegab	   = dble(omega_b)                           !DE80
	omegac	   = dble(omega_c)                           !DE
	omegav	   = dble(omega_v)                           !DE
	omegan	   = dble(omega_n)                           !DE
        omegaphif  = dble(omega_p)                           !DE
        wv         = dble(w_v)                               !DE
        wvbeg      = dble(w_vbeg)                            !DE
        wphif      = dble(w_p)                               !DE
        alphapot=0.                                          !DE
        amplipot=0.                                          !DE
        phif0=1.                                             !DE
	ipot       = ipot_p                                  !DE
	if((ipot.eq.1).or.(ipot.eq.2))then                   !DE
cDE     Setting initial potential exponent -2, to be adjusted numerically to match 
cDE     the equation of state today
        alphapot=-2. 								                      !DE
cDE     Setting initial potential amplitude at the Planck mass, to be fixed numerically later to match 
cDE     the required Omega_phi today
        amplipot=1.							                              !DE
cDE     Setting initial scalar field value at one tenth of the Planck mass, to be fixed numerically later 
cDE     to match the required Omega_phi today 
        phif0=0.1		                                                                      !DE
cDE	Setting initial ratio between kinetic and potential energy densities equal to 0
cDE	Tracking trajectories are greatly stable with respect to variation of this parameter
	rkinpot=0.                                                                                    !DE
cDE        write (*,*)'Enter initial kinetic versus potential energy ratio [0.]:'                     !DE
cDE        read (*,*) rkinpot                                                                         !DE
cDE     reading file names for background and perturbation amplitude
        fileprt='prt.dat'                                                                             !DE
cDE	Setting default name for background dynamics
cDE        print*,'Enter name of file with background dynamics 
cDE     &  in scalar field dark energy models [bkg.dat]:'                                             !DE
cDE        read(*,*) filebkg                                                                          !DE
	filebkg='bkg.dat'                                                                             !DE
	else if(ipot.ge.3) then                                                                       !DE
        print*,'sorry, only tracking inverse power law or SUGRA supported.'                           !DE
        stop                                                                                          !DE
	end if                                                                                        !DE
        betanmc=0.                                           !DE
        cnmc=0.                                              !DE
        omgjbd=1.d15                                         !DE
        sgnjbd=1.                                            !DE
        inmc       = inmc_p                                  !DE
	if(inmc.eq.1)then                                                                             !DE
cDE     Setting coupling exponent 2
        betanmc=2. 								                      !DE
        omgjbd     = dble(omgjbd_p)                          !DE
	sgnjbd     = dble(sgnjbd_p)                          !DE
cDE     Setting initial value for the non-minimal coupling constant, to be adjusted numerically later to
cDE     match the required OmegaJBD today
        cnmc=sgnjbd/dsqrt(32.d0*pi*omgjbd*phif0**2)                                                   !DE
	else if(inmc.ge.2) then                                                                       !DE
        print*,'sorry, only non-minimal coupling supported'                                           !DE
        stop                                                                                          !DE
	end if                                                                                        !DE
	omega=omegab+omegac+omegav+omegan+omegaphif                                                   !DE
	omegak=1.0d0-omega

cDE     Scalar field dark energy parameters
cDE     the generalised Lagrangian is (1/2)fnmc(phif)R-(1/2)onmc(phif)*phif_;mu*phif^;mu-V(phif)+L_matter ;
cDE     ipot=0 nullifies the potential, V=0 ;
cDE     inmc=0 nullifies non-minimal coupling, fnmc=0, onmc=1 ;
cDE     ipot=1 makes inverse power law potential V(phif)=M^4/(phif/phif0)^alphapot and field in tracking regime ;
cDE     ipot=2 makes SUGRA potential V(phif)=(M^4/(phif/phif0)^alphapot)exp(4pi(phif^2-phif0^2)) and field in tracking regime ;
cDE     inmc=1 makes fnmc=(1+8pi*cnmc*(phif^betanmc-phif0^betanmc)*R/(8piG), today f(phif0)=1/(8piG), onmc=1 ;
cDE     now initializes variables: 

1968    write (*,*) 'Enter scalar field dark energy non-minimal coupling index 
     &  [0 for minimal coupling, 1/8piG, 1 for 1/8piG+xi*phi^2]:'                                     !DE
        read (*,*) inmc                                                                               !DE

	        else if(omegaphif.eq.0.0)then                                                         !DE
        wphif=-1.0                                                                                    !DE
		else if(omegaphif.lt.0.)then                                                          !DE
	write(*,*) 'Sorry, only positive densities 
     &  for scalar field are allowed.'                                                                !DE
	stop                                                                                          !DE
		end if                                            


cDE     avoid omegak and omegaphi models 
        if (abs(omegak).ge.1.e-3.and.omegaphif.ne.0.0) then                           !DE90
        write(*,*) 'scalar field dark energy is allowed only in flat universe'        !DE
        stop                                                                          !DE
        end if                                                                        !DE

 2	continue

	h0	= dble(H_0)
	tcmb	= dble(t_cmb)
	yhe	= dble(Y_he)
	annur	= dble(a_nu_r)
	annunr	= dble(a_nu_nr)

        akmaxt=akmaxt*(h0/100.0d0)
	if (tcmb.lt.2.7d0.or.tcmb.gt.2.8d0) then
	  write(*,*)
     2      '  Warning: Tcmb has units of K.  Your value is weird.'
	end if

	if (yhe.lt.0.2d0.or.yhe.gt.0.3d0) then
	  write(*,*)
     2      '  Warning: Y_He is the Helium fraction of baryons.',
     3      '  Your value is weird.'
	end if
        if (annunr.lt.0.or.annunr.gt.3.1) then
          write(*,*)
     2      'Warning: N_nu(massive) is strange'
          write(*,*) '  Illegal choice.  Try again:'
          go to 2
        end if
        if (annur.lt.0.or.annur.gt.3.1) then
          write(*,*)
     2      'Warning: N_nu(massless) is strange'
          write(*,*) '  Illegal choice.  Try again:'
          go to 2
        end if
        if (annunr.lt.1.and.omegan.gt.0.0) then
          write(*,*)
     2      'Warning: N_nu(massive) should be 1, 2, or 3',
     3      'For non cero omegan'
	   write(*,*) '  Illegal choice.  Try again:'
          go to 2
        end if

cDE     Peebles recombination: required for dark energy models
	rflag=0
        if (rcflag.eq.1) then 
	   write(*,*) 'Calling recfast'
	   call recfast(omegab,omegac,omegav,h0,tcmb,yhe)
	   write(*,*)'recfast succesfully completed'
        endif

	if (tau.eq.0.) then 
	  riflag = 0
	else 
	  riflag = 1
	end if


c  Evaluate general cosmology constants.

c  grho gives the contribution to the expansion rate from: (g) photons,
c  (r) one flavor of relativistic neutrino (2 degrees of freedom),
c  (m) nonrelativistic matter (for Omega=1).  grho is actually
c  8*pi*G*rho/c^2 at a=1, with units of Mpc**(-2).
c  a=tau(Mpc)*adotrad, with a=1 today, assuming 3 neutrinos.
c  (Used only to set the initial conformal time.)
	grhom=3.3379d-11*h0*h0
	grhog=1.4952d-13*tcmb**4
	grhor=3.3957d-14*tcmb**4

c  adotrad gives the relation a(tau) in the radiation era:
	adotrad=2.8948d-7*tcmb*tcmb

	
	zri=0.0d0
	rif=0.0d0
	optdlss=0.0d0
	if (riflag.eq.1) then 
cDE     optical depth
	   optdlss = tau
	   rif=1.0d0
	end if

	if (riflag.eq.2) then
	   write (*,*)'Enter redshift, ionization fraction(e.g. 50 0.2)'
	   read (*,*) zri,rif
	   zristp=0.07d0*zri-1.0d0
	   if (zristp.lt.0.0) zristp=0.0d0
	end if

cDE	cosmological perturbations:
	if (ict.ne.1) then
	      if (type.eq.0) then 
	        itflag = 0
	      else 
	        itflag = 2
	      end if

cDE	scalar spectral indeces:
	   if (itflag.ne.2) then	 
	      nn=num_slope

	      do i=1,num_slope
		 an(i)=slope_n(i)
	      end do
	   end if
	   if (itflag.eq.1) then 
	      write(*,*)
     2 'Tensor spectral index given by nt=ns-1 (0) or different (1)'
	      read(*,*)itn
	      if (itn.eq.0) then
		 do 123 in=1,nn
		    ant(in)=an(in)-1
 123		 continue
	      else
		 write(*,*)'values of tensor spectral indexes:'
		 read(*,*)(ant(in),in=1,nn)
	      endif
	   endif
	   if (itflag.eq.2) then

cDE	tensor spectral indeces:
	      nn=num_slope
	      do i=1,num_slope
		 ant(i)=slope_n(i)
	      end do

	      do in=1,nn
		 rat(in)=1.0d0
	      enddo
	   end if
           if (itflag.eq.1) then
	      write(*,*)'ratio of tensor to scalar quadrupole given by' 
	      write(*,*)'7(1-n_S) (0) or different (1):'
	      read(*,*)irt
	      if (irt.eq.0) then
		 do in=1,nn
		    rat(in)=7.0d0*(1.0d0-an(in))
		 enddo
	      else
		 write(*,*)'values of T/S:'
		 read(*,*)(rat(in),in=1,nn)
	      endif
	   end if
	   if (itflag.eq.0) then
	      do in=1,nn
		 rat(in)=1.0d0
	      enddo
           endif 
	   
	   lensflag=0
	   if ((itflag.ne.2).and.(ict.eq.2)) then
	      write(*,*)'Enter 0 unlensed Cls only,'
	      write(*,*)'1 lensed Cls, linear evolution,'
	      write(*,*)'2 lensed Cls, non-linear evolution [1]'                                      !DE95
	      if (itflag.eq.1) then
		 write(*,*)'ONLY SCALAR Cls ARE LENSED'
	      end if
	      read(*,*)lensflag
	   end if
	   if ((lensflag.ne.0).and.(ict.eq.0)) then
	      write(*,*)'You did not request the transfer function'
	      write(*,*)'calculation needed to do the lensing'
	      write(*,*)'you will have to start again'
	      stop
	   else
c The last requested transfer function will be used for the lensing	      
	      ntlens=ntf
	   end if

cDE        check way of writing outputs 
	   write(*,*) 'Please, enter 1 for sharp writing, 0 otherwise: ' !DE100
           read(*,*) iwriteway                                           !DE
	   if (iwriteway.eq.0) then 	                                 !DE
	   if (itflag.ne.2) then
c	      write(*,*)'Enter output filename for SCALAR cl'
	      if (lensflag.ne.0) then
		 write(*,*)'If lensing was requested this will'
		 write(*,*)'be the unlensed power spectrum'
	      end if

c	      read(*,'(a80)')filename
	      filename = outfile

	      open(unit=8,file=filename,
     &	      	   status='unknown',form='formatted')
	      rewind 8

	      if (lensflag.ne.0) then
		 write(*,*)'Enter output filename for LENSED SCALAR cl'
		 read(*,'(a80)')filename
		 open(unit=10,file=filename,
     &            status='unknown',form='formatted')
		 rewind 10

		 write(*,*)'Enter file with bessel functions' 
                 write(*,*)' for lensing calculation (jlens.dat)'
		 read(*,'(a80)')filejlens
	      end if
	   end if

	   if (itflag.ne.0) then
cDE	output file name for tensors 
	      filename = outfile
	      open(unit=9,file=filename,
     &      	   status='unknown',form='formatted')
	      rewind 9
	   end if
	  else if (iwriteway.ne.0) then               	     !DE110
	      write(*,*) 'Sharp writing mode selected.'      !DE
	      write(*,*)'Enter output filename for cl'       !DE 
	      read(*,'(a80)')filename                        !DE 
	      open(unit=8,file=filename,                     !DE 
     &	      	   status='unknown',form='formatted')        !DE
	      rewind 8                                       !DE
          end if                                             !DE

	else
	   itflag=0
	   nn=0
	   lensflag=0
	end if

	if (itflag.ne.2) then
	   if (abs(omegak).gt.1.0d-3) then
	      initfl = 1
	   else
	      initfl = 1
	   end if
	else
	   initfl=1
	end if

c     TIMING
c	timeprev=actual
c	actual=etime(tarray)
c	write(50,*)actual-timeprev,' END INPUT, NOW MAIN'
c	write(50,*)'PARAMETERS'
c	write(50,*)omegab,omegac,omegav,wv,wvbeg,omegan,omegaphif,wphif !DE115
c	write(50,*)h0,tcmb,yhe,annur,annunr
c	write(50,*)riflag,zri,rif,optdlss
c	write(*,*)actual-timeprev,' END INPUT, NOW MAIN'
c	write(*,*)'PARAMETERS'
c	write(*,*)omegab,omegac,omegav,wv,wvbeg,omegan,omegaphif,wphif  !DE116
c	write(*,*)h0,tcmb,yhe,annur,annunr
c	write(*,*)riflag,zri,rif,optdlss
c     TESTING INPUT cosmological parameters.
c	write(123,*)omegab,omegac,omegav,omegan,omegak,h0,
c     &                     tcmb,yhe,annur,annunr

c call main subroutine

CC2 If you are interested in constructing a driver for
CC2 only the flat or open models code, leave the appropiate 
CC2 piece of following if statement. Then enter the appropiate
CC2 parameters for that case. 

	do in=1,nnmax
	   do il=1,l0max
	      clts(il,in)=0.0d0
	      cltt(il,in)=0.0d0
	      cles(il,in)=0.0d0
	      clet(il,in)=0.0d0
	      clbs(il,in)=0.0d0
	      clbt(il,in)=0.0d0
	      clcs(il,in)=0.0d0
	      clct(il,in)=0.0d0
	   end do
	end do


       if (abs(omegak).le.1.0d-3) then
CC1 initjl reads the tables of spherical Bessel functions
CC1 so if you are making a loop over models you should
CC1 move this call outside the loop.
	  if (ict.ne.1) then
c	     write(*,*)'Enter input filename for jl'
c	     read(*,'(a80)')filename
	     filename='jl.dat'
	     call initjl(filename)
	  end if
          call cmbflat(clts,cltt,cles,clet,
     &    clbt,clcs,clct,clkk,cltk)
       else
	     if (ict.ne.1) then
c		write(*,*)'Enter input filename for ujl'
c		read(*,'(a80)')filename
		filename='ujl.dat'
		call initujl0(filename)
	     end if
             call cmbopen(clts,cltt,cles,clet,
     1	   clbt,clcs,clct,clkk,cltk)
          end if

c       TIMING
c	timeprev=actual
c	actual=etime(tarray)
c	write(50,*)actual-timeprev,' END MAIN, NOW COBE'
c	write(*,*)actual-timeprev,' END MAIN, NOW COBE'


c	call COBEnormalize(clts,cltt,cles,clet,clbt,
c     &  clcs,clct,clkk,cltk)

c       TIMING
c	timeprev=actual
c	actual=etime(tarray)
c	write(50,*)actual-timeprev,' END COBE, NOW OUTPUT'


	  call output(clts,cltt,cles,clet,clbt,
     &  clcs,clct,itflag,lmo)

c       TIMING
c	timeprev=actual
c	actual=etime(tarray)
c	write(50,*)actual-timeprev,' END OUTPUT, NOW LENSING'

	
	if (lensflag.ne.0) then
	   call lensing(clts,cles,clbs,clcs,clkk,cltk)
	end if

	if (itflag.ne.2) close(8)
	if (itflag.ne.0) close(9)

c       TIMING
c	timeprev=actual
c	actual=etime(tarray)
c	write(50,*)actual-timeprev,' END LENSING, NOW END'
c	write(*,*)actual-timeinit,'sec = TOTAL TIME'

CC1 End your Loop over models here.
       return
       end
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc





